@@ -1,4 +1,4 @@
-/* $Id: plotX.c,v 1.6.2.2 2003/11/03 13:37:09 bill Exp $
+/* $Id: plotX.c,v 1.9 2002/06/09 18:49:11 karim Exp $
 
 Copyright (C) 2000  The PARI group.
 
@@ -18,6 +18,16 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
 /*                       HIGH RESOLUTION PLOT                      */
 /*                                                                 */
 /*******************************************************************/
+
+#ifdef BOTH_GNUPLOT_AND_X11		/* The switch support in plotgnuplot */
+#  define rectdraw0		X11_rectdraw0
+#  define term_set		X11_term_set
+#  define PARI_get_plot		X11_PARI_get_plot
+#  define plot_outfile_set	X11_plot_outfile_set
+#  define set_pointsize		X11_set_pointsize
+#  define pari_plot		pari_X11plot
+#endif
+
 #include "pari.h"
 #include "rect.h"
 #include "../language/anal.h"
@@ -135,7 +145,7 @@ rectdraw0(long *w, long *x, long *y, long lw, long do_free)
   long *numpoints[MAX_COLORS],*numtexts[MAX_COLORS];
   long *xtexts[MAX_COLORS],*ytexts[MAX_COLORS];
   long rcolcnt[MAX_COLORS][ROt_MAX];
-  long col,i,j,x0,y0,oldwidth,oldheight;
+  long col,i,j,x0,y0,a,b,oldwidth,oldheight;
   long rcnt[ROt_MAX+1], hjust, vjust, hgap, vgap, hgapsize, vgapsize;
   char **texts[MAX_COLORS];
   PariRect *e;
@@ -159,6 +169,9 @@ rectdraw0(long *w, long *x, long *y, long lw, long do_free)
 
   /* child process goes on */
   freeall();  /* PARI stack isn't needed anymore, keep rectgraph */
+  /* if gnuplot X11 plotting is active, may get SIGPIPE...  XXXX Better disable
+   * some X callback? */
+  os_signal(SIGPIPE, SIG_IGN);
   PARI_get_plot(1);
   display = XOpenDisplay(NULL);
   font_info = XLoadQueryFont(display, "9x15");
@@ -307,10 +320,10 @@ rectdraw0(long *w, long *x, long *y, long lw, long do_free)
 		seg[col][c[ROt_LN]].y2 = DTOL((RoLNy2(p1)+y0)*ys);
 		c[ROt_LN]++;break;
 	      case ROt_BX:
-		rec[col][c[ROt_BX]].x = DTOL((RoBXx1(p1)+x0)*xs);
-		rec[col][c[ROt_BX]].y = DTOL((RoBXy1(p1)+y0)*ys);
-		rec[col][c[ROt_BX]].width = DTOL((RoBXx2(p1)-RoBXx1(p1))*xs);
-		rec[col][c[ROt_BX]].height = DTOL((RoBXy2(p1)-RoBXy1(p1))*ys);
+		a=rec[col][c[ROt_BX]].x = DTOL((RoBXx1(p1)+x0)*xs);
+		b=rec[col][c[ROt_BX]].y = DTOL((RoBXy1(p1)+y0)*ys);
+		rec[col][c[ROt_BX]].width = DTOL((RoBXx2(p1)+x0-a)*xs);
+		rec[col][c[ROt_BX]].height = DTOL((RoBXy2(p1)+y0-b)*ys);
 		c[ROt_BX]++;break;
 	      case ROt_MP:
 		ptx = RoMPxs(p1); pty = RoMPys(p1);
